﻿using System;                           /* for types (Byte, IntPtr, etc.) */
using System.IO.Ports;                  /* for the RS232 interface */
using System.Threading;                 /* for Sleep() */
using System.Runtime.InteropServices;   /* for P/Invoke (native interface) */


public interface TwinklerIO
{
    bool OpenPort(string PortName, int BaudRate);
    void ClosePort();
    bool SendData(byte[] Data, int DataSize);
    bool SendBreak();
    byte BitBang(byte Mask, byte Bits);
}


public class TwinklerCom : TwinklerIO
{
    SerialPort ComPort;

    public bool OpenPort(string PortName, int BaudRate)
    {
        ComPort = new SerialPort(PortName, BaudRate, Parity.None, 8, StopBits.One);
        ComPort.Handshake = Handshake.None;
        try {
            ComPort.Open();
            return ComPort.IsOpen;
        } catch {
            return false;
        } /* try */
    }

    public void ClosePort()
    {
        if (ComPort.IsOpen)
            ComPort.Close();
    }

    public bool SendData(byte[] Data, int DataSize)
    {
        if (ComPort.IsOpen)
            ComPort.Write(Data, 0, DataSize);
        return ComPort.IsOpen;
    }

    public bool SendBreak()
    {
        if (!ComPort.IsOpen)
            return false;
        ComPort.BreakState = true;
        Thread.Sleep(200);
        ComPort.BreakState = false;
        return true;
    }

    public byte BitBang(byte Mask, byte Bits)
    {
        return 0xff;
    }
}


public class TwinklerNative : TwinklerIO
{
    static IntPtr INVALID_HANDLE_VALUE = new IntPtr(-1);

    const UInt32 FT_OK = 0;

    const Byte FT_BITS_8 = 8;
    const Byte FT_STOP_BITS_1 = 0;
    const Byte FT_PARITY_NONE = 0;
    const UInt16 FT_FLOW_NONE = 0;

    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_Open(int DeviceNumber, out IntPtr ftHandle);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_ResetDevice(IntPtr ftHandle);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_SetBaudRate(IntPtr ftHandle, UInt32 BaudRate);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_SetDataCharacteristics(IntPtr ftHandle, Byte WordLength, Byte StopBits, Byte Parity);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_SetFlowControl(IntPtr ftHandle, UInt16 FlowControl, Byte XonChar, Byte XoffChar);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_Close(IntPtr ftHandle);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_Write(IntPtr ftHandle, Byte[] Buffer, UInt32 BytesToWrite, out UInt32 BytesWritten);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_SetBreakOn(IntPtr ftHandle);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_SetBreakOff(IntPtr ftHandle);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_SetBitMode(IntPtr ftHandle, Byte Mask, Byte Enable);
    [DllImport("ftd2xx.dll")]
    static extern UInt32 FT_GetBitMode(IntPtr ftHandle, out Byte Mode);


    IntPtr hFT;

    TwinklerNative()
    {
        hFT = INVALID_HANDLE_VALUE;
    }

    ~TwinklerNative()
    {
        ClosePort();
    }

    public bool OpenPort(string PortName, int BaudRate)
    {
        int index = System.Convert.ToInt32(PortName);
        if (FT_Open(index, out hFT) == FT_OK) {
            FT_ResetDevice(hFT);
            FT_SetBaudRate(hFT, (UInt32)BaudRate);
            FT_SetDataCharacteristics(hFT, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE);
            FT_SetFlowControl(hFT, FT_FLOW_NONE, 0, 0);
        } else {
            hFT = INVALID_HANDLE_VALUE;
        } /* if */
        return hFT != INVALID_HANDLE_VALUE;
    }

    public void ClosePort()
    {
        if (hFT != INVALID_HANDLE_VALUE) {
            FT_Close(hFT);
            hFT = INVALID_HANDLE_VALUE;
        } /* if */
    }

    public bool SendData(byte[] Data, int DataSize)
    {
        uint written = 0;
        if (hFT == INVALID_HANDLE_VALUE || FT_Write(hFT, Data, (UInt32)DataSize, out written) != FT_OK)
            written = 0;
        return written == (uint)DataSize;
    }

    public bool SendBreak()
    {
        if (hFT == INVALID_HANDLE_VALUE)
            return false;
        FT_SetBreakOn(hFT);
        Thread.Sleep(200);
        FT_SetBreakOff(hFT);
        return true;
    }

    public byte BitBang(byte Mask, byte Bits)
    {
        if (hFT == INVALID_HANDLE_VALUE)
            return 0;
        byte result = 0;
        byte msk = (byte)((Mask << 4) | Bits);
        FT_SetBitMode(hFT, msk, 0x20);
        FT_GetBitMode(hFT, out result);
        return (byte)(result & 0x0f);
    }
}


public class Twinkler
{
    TwinklerIO Port;
    byte cbus;

    /* ----- general constants and routines ----- */
    public const byte Twinkler_ColourArray = 0xF0;
    public const byte Twinkler_SetColour = 0xF1;
    public const byte Twinkler_FadePeriod = 0xF2;
    public const byte Twinkler_BlinkRate = 0xF3;
    public const byte Twinkler_SetState = 0xF4;
    public const byte Twinkler_SetRange = 0xF5;
    public const byte Twinkler_SetBaud = 0xF6;
    public const byte Twinkler_StoreState = 0xF7;
    public const byte Twinkler_TickMark = 0xF8;

    public static byte Twinkler_RGB(int red, int green, int blue)
    {
        return (byte)(36 * red + 6 * green + blue);
    }

    /* ----- constructor/destructor ----- */
    public Twinkler(TwinklerIO port)
    {
        Port = port;
    }

    ~Twinkler()
    {
        Port.ClosePort();
    }

    /* ----- forwarded low-level routines ----- */
    public bool OpenPort(string PortName, int BaudRate)
    {
        return Port.OpenPort(PortName, BaudRate);
    }

    public void ClosePort()
    {
        Port.ClosePort();
    }

    public bool SendData(byte[] Data, int DataSize)
    {
        return Port.SendData(Data, DataSize);
    }

    public bool Reset()
    {
        return Port.SendBreak();
    }

    public byte BitBang(byte Mask, byte Bits)
    {
        return Port.BitBang(Mask, Bits);
    }


    /* ----- high-level functions ----- */

    public bool ColourArray(byte[] Data, int DataSize)
    {
        byte[] cmd = { Twinkler_ColourArray };
        Port.SendData(cmd, 1);
        return Port.SendData(Data, DataSize);
    }

    public bool SetColour(byte Colour)
    {
        byte[] cmd = { Twinkler_SetColour, Colour };
        return Port.SendData(cmd, 2);
    }

    public bool SetColour(byte Red, byte Green, byte Blue)
    {
        byte[] cmd = { Twinkler_SetColour, Twinkler_RGB(Red, Green, Blue) };
        return Port.SendData(cmd, 2);
    }

    public bool FadePeriod(byte Period)
    {
        byte[] cmd = { Twinkler_FadePeriod, Period };
        return Port.SendData(cmd, 2);
    }

    public bool BlinkRate(byte Rate)
    {
        byte[] cmd = { Twinkler_BlinkRate, Rate };
        return Port.SendData(cmd, 2);
    }

    public bool SetState(byte State)
    {
        byte[] cmd = { Twinkler_SetState, State };
        return Port.SendData(cmd, 2);
    }

    public bool SetRange(int First, int Last)
    {
        byte[] cmd = new byte[5];
        cmd[0] = Twinkler_SetRange;
        cmd[1] = (byte)(First & 0x7f);
        cmd[2] = (byte)((First >> 7) & 0x7e);
        cmd[3] = (byte)((Last - First) & 0x7f);
        cmd[4] = (byte)(((Last - First) >> 7) & 0x7e);
        return Port.SendData(cmd, 5);
    }

    public bool SetRange()
    {
        return SetRange(0, 16255);
    }

    public bool SetBaud(byte Baud)
    {
        byte[] cmd = { Twinkler_SetBaud, Baud };
        return Port.SendData(cmd, 2);
    }

    public bool StoreState(byte Hold, byte State)
    {
        byte[] cmd = { Twinkler_StoreState, Hold, State };
        return Port.SendData(cmd, 3);
    }

    public bool TickMark()
    {
        byte[] cmd = { Twinkler_TickMark };
        return Port.SendData(cmd, 1);
    }

    public byte ReadInputs()
    {
    	byte sw = Port.BitBang(0, 0);
    	if (sw != cbus) {
    		/* debounce */
    		int count = 0;
    		while (count < 5) {
    			cbus = sw;
                Thread.Sleep(5);
    			sw = Port.BitBang(0, 0);
    			count = (cbus == sw) ? count + 1 : 0;
    		} /* while */
    	} /* if */
        return sw;
    }
}
